# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from typing import Any, Dict, List, Tuple

import gi
from gi.repository import GLib
from dbus.mainloop.glib import DBusGMainLoop

import dbus
import dbusmock
import fcntl
import logging
import os
import pytest
import subprocess

logging.basicConfig(format="%(levelname)s | %(name)s: %(message)s", level=logging.DEBUG)
logger = logging.getLogger("pyportaltest")

DBusGMainLoop(set_as_default=True)

# Uncomment this to have dbus-monitor listen on the normal session address
# rather than the test DBus. This can be useful for cases where *something*
# messes up and tests run against the wrong bus.
#
# session_dbus_address = os.environ["DBUS_SESSION_BUS_ADDRESS"]


def start_dbus_monitor() -> "subprocess.Process":
    import subprocess

    env = os.environ.copy()
    try:
        env["DBUS_SESSION_BUS_ADDRESS"] = session_dbus_address
    except NameError:
        # See comment above
        pass

    argv = ["dbus-monitor", "--session"]
    mon = subprocess.Popen(argv, env=env)

    def stop_dbus_monitor():
        mon.terminate()
        mon.wait()

    GLib.timeout_add(2000, stop_dbus_monitor)
    return mon


class PortalTest(dbusmock.DBusTestCase):
    """
    Parent class for portal tests. Subclass from this and name it after the
    portal, e.g. ``TestWallpaper``.

    .. attribute:: portal_interface

        The :class:`dbus.Interface` referring to our portal

    .. attribute:: properties_interface

        A convenience :class:`dbus.Interface` referring to the DBus Properties
        interface, call ``Get``, ``Set`` or ``GetAll`` on this interface to
        retrieve the matching property/properties.

    .. attribute:: mock_interface

        The DBusMock :class:`dbus.Interface` that controls our DBus
        appearance.

    """

    @classmethod
    def setUpClass(cls):
        if cls.__name__ != "PortalTest":
            cls.PORTAL_NAME = cls.__name__.removeprefix("Test")
            cls.INTERFACE_NAME = f"org.freedesktop.portal.{cls.PORTAL_NAME}"
        os.environ["LIBPORTAL_TEST_SUITE"] = "1"

        try:
            dbusmock.mockobject.DBusMockObject.EmitSignalDetailed
        except AttributeError:
            pytest.skip("Updated version of dbusmock required")

        cls.__have_session_bus = False

    @classmethod
    def ensure_session_bus(cls):
        if not cls.__have_session_bus:
            cls.__have_session_bus = True
            cls.start_session_bus()

    def setUp(self):
        self.p_mock = None
        self._mainloop = None
        self.dbus_monitor = None

    def setup_daemon(self, params=None, extra_templates: List[Tuple[str, Dict]] = []):
        """
        Start a DBusMock daemon in a separate process.

        If extra_templates is specified, it is a list of tuples with the
        portal name as first value and the param dict to be passed to that
        template as second value, e.g. ("ScreenCast", {...}).
        """
        self.ensure_session_bus()
        self.p_mock, self.obj_portal = self.spawn_server_template(
            template=f"pyportaltest/templates/{self.PORTAL_NAME.lower()}.py",
            parameters=params,
            stdout=subprocess.PIPE,
        )

        flags = fcntl.fcntl(self.p_mock.stdout, fcntl.F_GETFL)
        fcntl.fcntl(self.p_mock.stdout, fcntl.F_SETFL, flags | os.O_NONBLOCK)
        self.mock_interface = dbus.Interface(self.obj_portal, dbusmock.MOCK_IFACE)
        self.properties_interface = dbus.Interface(
            self.obj_portal, dbus.PROPERTIES_IFACE
        )
        self.portal_interface = dbus.Interface(self.obj_portal, self.INTERFACE_NAME)

        for t, tparams in extra_templates:
            template = f"pyportaltest/templates/{t.lower()}.py"
            self.obj_portal.AddTemplate(
                template,
                dbus.Dictionary(tparams, signature="sv"),
                dbus_interface=dbusmock.MOCK_IFACE,
            )

        self.dbus_monitor = start_dbus_monitor()

    def tearDown(self):
        if self.p_mock:
            if self.p_mock.stdout:
                out = (self.p_mock.stdout.read() or b"").decode("utf-8")
                if out:
                    print(out)
                self.p_mock.stdout.close()
            self.p_mock.terminate()
            self.p_mock.wait()

        if self.dbus_monitor:
            self.dbus_monitor.terminate()
            self.dbus_monitor.wait()

    @property
    def mainloop(self):
        """
        The mainloop for this test. This mainloop automatically quits after a
        fixed timeout, but only on the first run. That's usually enough for
        tests, if you need to call mainloop.run() repeatedly ensure that a
        timeout handler is set to ensure quick test case failure  in case of
        error.
        """
        if self._mainloop is None:

            def quit():
                self._mainloop.quit()
                self._mainloop = None

            self._mainloop = GLib.MainLoop()
            GLib.timeout_add(2000, quit)

        return self._mainloop

    def assert_version_eq(self, version: int):
        """Assert the given version number is the one our portal exports"""
        interface_name = self.INTERFACE_NAME
        params = {}
        self.setup_daemon(params)
        assert self.properties_interface.Get(interface_name, "version") == version



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from pyportaltest.templates import Request, Response, Session, ASVType, MockParams
from typing import Dict, List, Tuple, Iterator
from itertools import count

import dbus
import dbus.service
import logging
import socket

logger = logging.getLogger(f"templates.{__name__}")

BUS_NAME = "org.freedesktop.portal.Desktop"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.portal.RemoteDesktop"

_restore_tokens = count()


def load(mock, parameters):
    logger.debug(f"loading {MAIN_IFACE} template with params {parameters}")

    params = MockParams.get(mock, MAIN_IFACE)
    params.delay = 500
    params.version = parameters.get("version", 2)
    params.response = parameters.get("response", 0)
    params.devices = parameters.get("devices", 0b111)
    params.sessions: Dict[str, Session] = {}
    params.close_after_start = parameters.get("close-after-start", 0)

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "version": dbus.UInt32(params.version),
                "AvailableDeviceTypes": dbus.UInt32(
                    parameters.get("device-types", params.devices)
                ),
            }
        ),
    )


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="a{sv}",
    out_signature="o",
)
def CreateSession(self, options, sender):
    try:
        logger.debug(f"CreateSession: {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        session = Session(bus_name=self.bus_name, sender=sender, options=options)
        params.sessions[session.handle] = session

        response = Response(params.response, {})

        request.respond(response, delay=params.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}",
    out_signature="o",
)
def SelectDevices(self, session_handle, options, sender):
    try:
        logger.debug(f"SelectDevices: {session_handle} {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        response = Response(params.response, {})
        request.respond(response, delay=params.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="osa{sv}",
    out_signature="o",
)
def Start(self, session_handle, parent_window, options, sender):
    try:
        logger.debug(f"Start: {session_handle} {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        results = {
            "devices": dbus.UInt32(params.devices),
        }

        response = Response(params.response, results)

        request.respond(response, delay=params.delay)

        if params.close_after_start > 0:
            session = params.sessions[session_handle]
            session.close({}, params.close_after_start)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}dd",
    out_signature="",
)
def NotifyPointerMotion(self, session_handle, options, dx, dy):
    try:
        logger.debug(f"NotifyPointerMotion: {session_handle} {options} {dx} {dy}")
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}udd",
    out_signature="",
)
def NotifyPointerMotionAbsolute(self, session_handle, options, stream, x, y):
    try:
        logger.debug(
            f"NotifyPointerMotionAbsolute: {session_handle} {options} {stream} {x} {y}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}iu",
    out_signature="",
)
def NotifyPointerButton(self, session_handle, options, button, state):
    try:
        logger.debug(
            f"NotifyPointerButton: {session_handle} {options} {button} {state}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}dd",
    out_signature="",
)
def NotifyPointerAxis(self, session_handle, options, dx, dy):
    try:
        logger.debug(f"NotifyPointerAxis: {session_handle} {options} {dx} {dx}")
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}ui",
    out_signature="",
)
def NotifyPointerAxisDiscrete(self, session_handle, options, axis, steps):
    try:
        logger.debug(
            f"NotifyPointerAxisDiscrete: {session_handle} {options} {axis} {steps}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}iu",
    out_signature="",
)
def NotifyKeyboardKeycode(self, session_handle, options, keycode, state):
    try:
        logger.debug(
            f"NotifyKeyboardKeycode: {session_handle} {options} {keycode} {state}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}iu",
    out_signature="",
)
def NotifyKeyboardKeysym(self, session_handle, options, keysym, state):
    try:
        logger.debug(
            f"NotifyKeyboardKeysym: {session_handle} {options} {keysym} {state}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}uudd",
    out_signature="",
)
def NotifyTouchDown(self, session_handle, options, stream, slot, x, y):
    try:
        logger.debug(
            f"NotifyTouchDown: {session_handle} {options} {stream} {slot} {x} {y}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}uudd",
    out_signature="",
)
def NotifyTouchMotion(self, session_handle, options, stream, slot, x, y):
    try:
        logger.debug(
            f"NotifyTouchMotion: {session_handle} {options} {stream} {slot} {x} {y}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}u",
    out_signature="",
)
def NotifyTouchUp(self, session_handle, options, slot):
    try:
        logger.debug(f"NotifyTouchMotion: {session_handle} {options} {slot}")
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}",
    out_signature="h",
)
def ConnectToEIS(self, session_handle, options):
    try:
        logger.debug(f"ConnectToEIS: {session_handle} {options}")
        import socket

        sockets = socket.socketpair()
        # Write some random data down so it'll break anything that actually
        # expects the socket to be a real EIS socket
        sockets[0].send(b"VANILLA")
        fd = sockets[1]
        logger.debug(f"ConnectToEIS with fd {fd.fileno()}")
        return dbus.types.UnixFd(fd)
    except Exception as e:
        logger.critical(e)



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from . import PortalTest

import gi
import logging

gi.require_version("Xdp", "1.0")
from gi.repository import GLib, Xdp

logger = logging.getLogger(__name__)


class TestWallpaper(PortalTest):
    def test_version(self):
        self.assert_version_eq(1)

    def set_wallpaper(
        self, uri_to_set: str, set_on: Xdp.WallpaperFlags, show_preview: bool
    ):
        params = {}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        assert xdp is not None

        flags = {
            "background": Xdp.WallpaperFlags.BACKGROUND,
            "lockscreen": Xdp.WallpaperFlags.LOCKSCREEN,
            "both": Xdp.WallpaperFlags.BACKGROUND | Xdp.WallpaperFlags.LOCKSCREEN,
        }[set_on]

        if show_preview:
            flags |= Xdp.WallpaperFlags.PREVIEW

        wallpaper_was_set = False

        def set_wallpaper_done(portal, task, data):
            nonlocal wallpaper_was_set
            wallpaper_was_set = portal.set_wallpaper_finish(task)
            self.mainloop.quit()

        xdp.set_wallpaper(
            parent=None,
            uri=uri_to_set,
            flags=flags,
            cancellable=None,
            callback=set_wallpaper_done,
            data=None,
        )

        self.mainloop.run()

        method_calls = self.mock_interface.GetMethodCalls("SetWallpaperURI")
        assert len(method_calls) == 1
        timestamp, args = method_calls.pop(0)
        parent, uri, options = args
        assert uri == uri_to_set
        assert options["set-on"] == set_on
        assert options["show-preview"] == show_preview

        assert wallpaper_was_set

    def test_set_wallpaper_background(self):
        self.set_wallpaper("https://background.nopreview", "background", False)

    def test_set_wallpaper_background_preview(self):
        self.set_wallpaper("https://background.preview", "background", True)

    def test_set_wallpaper_lockscreen(self):
        self.set_wallpaper("https://lockscreen.nopreview", "lockscreen", False)

    def test_set_wallpaper_lockscreen_preview(self):
        self.set_wallpaper("https://lockscreen.preview", "lockscreen", True)

    def test_set_wallpaper_both(self):
        self.set_wallpaper("https://both.nopreview", "both", False)

    def test_set_wallpaper_both_preview(self):
        self.set_wallpaper("https://both.preview", "both", True)

    def test_set_wallpaper_cancel(self):
        params = {"response": 1}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        assert xdp is not None

        flags = Xdp.WallpaperFlags.BACKGROUND

        wallpaper_was_set = False

        def set_wallpaper_done(portal, task, data):
            nonlocal wallpaper_was_set
            try:
                wallpaper_was_set = portal.set_wallpaper_finish(task)
            except GLib.GError:
                pass
            self.mainloop.quit()

        xdp.set_wallpaper(
            parent=None,
            uri="https://ignored.anyway",
            flags=flags,
            cancellable=None,
            callback=set_wallpaper_done,
            data=None,
        )

        self.mainloop.run()

        method_calls = self.mock_interface.GetMethodCalls("SetWallpaperURI")
        assert len(method_calls) == 1

        assert not wallpaper_was_set



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from pyportaltest.templates import Request, Response, ASVType, MockParams
from typing import Dict, List, Tuple, Iterator

import dbus
import dbus.service
import logging

logger = logging.getLogger(f"templates.{__name__}")

BUS_NAME = "org.freedesktop.portal.Desktop"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.portal.Wallpaper"


def load(mock, parameters):
    logger.debug(f"loading {MAIN_IFACE} template")

    params = MockParams.get(mock, MAIN_IFACE)
    params.delay = 500
    params.response = parameters.get("response", 0)

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary({"version": dbus.UInt32(parameters.get("version", 1))}),
    )


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="ssa{sv}",
    out_signature="o",
)
def SetWallpaperURI(self, parent_window, uri, options, sender):
    try:
        logger.debug(f"SetWallpaperURI: {parent_window}, {uri}, {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        response = Response(params.response, {})

        request.respond(response, delay=params.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from pyportaltest.templates import Request, Response, Session, ASVType, MockParams
from typing import Dict, List, Tuple, Iterator
from itertools import count

import dbus
import dbus.service
import logging
import socket

logger = logging.getLogger(f"templates.{__name__}")

BUS_NAME = "org.freedesktop.portal.Desktop"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.portal.ScreenCast"

_restore_tokens = count()


def load(mock, parameters):
    logger.debug(f"loading {MAIN_IFACE} template")

    params = MockParams.get(mock, MAIN_IFACE)
    params.delay = 500
    params.version = parameters.get("version", 4)
    params.response = parameters.get("response", 0)
    # streams returned in Start
    params.streams = parameters.get("streams", [])
    # persist_mode returned in Start
    params.persist_mode = parameters.get("persist-mode", 0)
    params.sessions: Dict[str, Session] = {}

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "version": dbus.UInt32(params.version),
                "AvailableSourceTypes": dbus.UInt32(
                    parameters.get("source-types", 0b111)
                ),
                "AvailableCursorModes": dbus.UInt32(
                    parameters.get("cursor-modes", 0b111)
                ),
            }
        ),
    )


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="a{sv}",
    out_signature="o",
)
def CreateSession(self, options, sender):
    try:
        logger.debug(f"CreateSession: {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        session = Session(bus_name=self.bus_name, sender=sender, options=options)
        params.sessions[session.handle] = session

        response = Response(params.response, {})

        request.respond(response, delay=params.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}",
    out_signature="o",
)
def SelectSources(self, session_handle, options, sender):
    try:
        logger.debug(f"SelectSources: {session_handle} {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        response = Response(params.response, {})

        request.respond(response, delay=params.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="osa{sv}",
    out_signature="o",
)
def Start(self, session_handle, parent_window, options, sender):
    try:
        logger.debug(f"Start: {session_handle} {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        results = {
            "streams": params.streams,
        }

        if params.version >= 4:
            results["persist_mode"] = dbus.UInt32(params.persist_mode)
            if params.persist_mode != 0:
                results["restore_token"] = f"restore_token{next(_restore_tokens)}"

        response = Response(params.response, results)

        request.respond(response, delay=params.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}",
    out_signature="h",
)
def OpenPipeWireRemote(self, session_handle, options, sender):
    try:
        logger.debug(f"OpenPipeWireRemote: {session_handle} {options}")

        # libportal doesn't care about the socket, so let's use something we
        # can easily check
        sockets = socket.socketpair()

        pw_socket = sockets[0]
        pw_socket.send(b"I AM GROO^WPIPEWIRE")

        fd = sockets[1]

        logger.debug(f"OpenPipeWireRemote with fd {fd.fileno()}")

        return dbus.types.UnixFd(fd)
    except Exception as e:
        logger.critical(e)



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from pyportaltest.templates import Request, Response, ASVType, MockParams
from typing import Dict, List, Tuple, Iterator

import dbus
import dbus.service
import logging

logger = logging.getLogger(f"templates.{__name__}")

BUS_NAME = "org.freedesktop.portal.Desktop"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.portal.Notification"


def load(mock, parameters):
    logger.debug(f"loading {MAIN_IFACE} template")

    params = MockParams.get(mock, MAIN_IFACE)
    params.delay = 500
    params.version = parameters.get("version", 2)
    params.response = parameters.get("response", 0)

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary({"version": dbus.UInt32(params.version)}),
    )


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="sa{sv}",
    out_signature="",
)
def AddNotification(self, id, notification, sender):
    try:
        logger.debug(f"AddNotification: {id}, {notification}")

    except Exception as e:
        logger.critical(e)



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

"""xdg desktop portals mock template"""

from pyportaltest.templates import Request, Response, ASVType, Session
from typing import Callable, Dict, List, Tuple, Iterator
from itertools import count

_restore_tokens = count()

import dbus
import dbus.service
import logging
import sys

from gi.repository import GLib

BUS_NAME = "org.freedesktop.portal.Desktop"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.portal.InputCapture"

logger = logging.getLogger(f"templates.{__name__}")
logger.setLevel(logging.DEBUG)

zone_set = None
eis_serial = None


def load(mock, parameters={}):
    logger.debug(f"Loading parameters: {parameters}")

    # Delay before Request.response, applies to all functions
    mock.delay: int = parameters.get("delay", 0)

    # EIS serial number, < 0 means "don't send a serial"
    eis_serial_start = parameters.get("eis-serial", 0)
    if eis_serial_start >= 0:
        global eis_serial
        eis_serial = count(start=eis_serial_start)

    # Zone set number, < 0 means "don't send a zone_set"
    zone_set_start = parameters.get("zone-set", 0)
    if zone_set_start >= 0:
        global zone_set
        zone_set = count(start=zone_set_start)
        mock.current_zone_set = next(zone_set)
    else:
        mock.current_zone_set = None

    # An all-zeroes zone means "don't send a zone"
    mock.current_zones = parameters.get("zones", ((1920, 1080, 0, 0),))
    if mock.current_zones[0] == (0, 0, 0, 0):
        mock.current_zones = None

    # second set of zones after the change signal
    mock.changed_zones = parameters.get("changed-zones", ((0, 0, 0, 0),))
    if mock.changed_zones[0] == (0, 0, 0, 0):
        mock.changed_zones = None

    # milliseconds until the zones change to the changed_zones
    mock.change_zones_after = parameters.get("change-zones-after", 0)

    # List of barrier ids to fail
    mock.failed_barriers = parameters.get("failed-barriers", [])

    # When to send the Activated signal (in ms after Enable), 0 means no
    # signal
    mock.activated_after = parameters.get("activated-after", 0)

    # Barrier ID that triggers Activated (-1 means don't add barrier id)
    mock.activated_barrier = parameters.get("activated-barrier", None)

    # Position tuple for Activated signal, None means don't add position
    mock.activated_position = parameters.get("activated-position", None)

    # When to send the Deactivated signal (in ms after Activated), 0 means no
    # signal
    mock.deactivated_after = parameters.get("deactivated-after", 0)

    # Position tuple for Deactivated signal, None means don't add position
    mock.deactivated_position = parameters.get("deactivated-position", None)

    # When to send the Disabled signal (in ms after Enabled), 0 means no
    # signal
    mock.disabled_after = parameters.get("disabled-after", 0)

    # How many ms to signal Session.Closed after Start
    mock.close_after_enable = parameters.get("close-after-enable", 0)

    # persist_mode returned in Start response (0 means none)
    mock.start_persist_mode = parameters.get("start-persist-mode", 0)

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "version": dbus.UInt32(parameters.get("version", 1)),
                "SupportedCapabilities": dbus.UInt32(
                    parameters.get("capabilities", 0xF)
                ),
            }
        ),
    )

    mock.active_sessions: Dict[str, Session] = {}


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="sa{sv}",
    out_signature="o",
)
def CreateSession(self, parent_window: str, options: ASVType, sender: str):
    try:
        request = Request(bus_name=self.bus_name, sender=sender, options=options)
        session = Session(bus_name=self.bus_name, sender=sender, options=options)

        response = Response(
            0,
            {
                "capabilities": dbus.UInt32(0xF, variant_level=1),
                "session_handle": dbus.ObjectPath(session.handle),
            },
        )
        self.active_sessions[session.handle] = session

        logger.debug(f"CreateSession with response {response}")
        request.respond(response, delay=self.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="a{sv}",
    out_signature="a{sv}",
)
def CreateSession2(self, options: ASVType, sender: str):
    try:
        session = Session(bus_name=self.bus_name, sender=sender, options=options)
        self.active_sessions[session.handle] = session

        results = {
            "session_handle": dbus.ObjectPath(session.handle),
        }

        logger.debug(f"CreateSession2 with results {results}")
        return results
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="osa{sv}",
    out_signature="o",
)
def Start(self, session_handle: str, parent_window: str, options: ASVType, sender: str):
    try:
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        if session_handle not in self.active_sessions:
            request.respond(Response(2, {}), delay=self.delay)
            return request.handle

        results = {}

        if self.start_persist_mode != 0:
            results["restore_token"] = f"restore_token{next(_restore_tokens)}"

        response = Response(response=0, results=results)

        logger.debug(f"Start with response {response}")
        request.respond(response, delay=self.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}",
    out_signature="h",
)
def ConnectToEIS(self, session_handle: str, options: ASVType, sender: str):
    try:
        import socket

        sockets = socket.socketpair()
        # Write some random data down so it'll break anything that actually
        # expects the socket to be a real EIS socket
        sockets[0].send(b"VANILLA")
        fd = sockets[1]
        logger.debug(f"ConnectToEIS with fd {fd.fileno()}")
        return dbus.types.UnixFd(fd)
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}",
    out_signature="o",
)
def GetZones(self, session_handle: str, options: ASVType, sender: str):
    try:
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        if session_handle not in self.active_sessions:
            request.respond(Response(2, {}, delay=self.delay))
            return request.handle

        zone_set = self.current_zone_set
        zones = self.current_zones

        results = {}
        if zone_set is not None:
            results["zone_set"] = dbus.UInt32(zone_set, variant_level=1)
        if zones is not None:
            results["zones"] = dbus.Array(
                [dbus.Struct(z, signature="uuii") for z in zones],
                signature="(uuii)",
                variant_level=1,
            )

        response = Response(response=0, results=results)

        logger.debug(f"GetZones with response {response}")
        request.respond(response, delay=self.delay)

        if self.change_zones_after > 0:

            def change_zones():
                global zone_set

                logger.debug("Changing Zones")
                opts = {"zone_set": dbus.UInt32(self.current_zone_set, variant_level=1)}
                self.current_zone_set = next(zone_set)
                self.current_zones = self.changed_zones
                self.EmitSignalDetailed(
                    "",
                    "ZonesChanged",
                    "oa{sv}",
                    [dbus.ObjectPath(session_handle), opts],
                    details={"destination": sender},
                )

            GLib.timeout_add(self.change_zones_after, change_zones)

            self.change_zones_after = 0  # Zones only change once

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}aa{sv}u",
    out_signature="o",
)
def SetPointerBarriers(
    self,
    session_handle: str,
    options: ASVType,
    barriers: List[ASVType],
    zone_set: int,
    sender: str,
):
    try:
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        if (
            session_handle not in self.active_sessions
            or zone_set != self.current_zone_set
        ):
            response = Response(2, {})
        else:
            results = {
                "failed_barriers": dbus.Array(
                    self.failed_barriers, signature="u", variant_level=1
                )
            }
            response = Response(0, results)

        logger.debug(f"SetPointerBarriers with response {response}")
        request.respond(response, delay=self.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}",
    out_signature="",
)
def Enable(self, session_handle, options, sender):
    try:
        logger.debug(f"Enable with options {options}")
        allowed_options = []

        if not all([k in allowed_options for k in options]):
            logger.error("Enable does not support options")

        if self.activated_after > 0:
            current_eis_serial = next(eis_serial) if eis_serial else None

            def send_activated():
                opts = {}
                if current_eis_serial is not None:
                    opts["activation_id"] = dbus.UInt32(
                        current_eis_serial, variant_level=1
                    )

                if self.activated_position is not None:
                    opts["cursor_position"] = dbus.Struct(
                        self.activated_position, signature="dd", variant_level=1
                    )
                if self.activated_barrier is not None:
                    opts["barrier_id"] = dbus.UInt32(
                        self.activated_barrier, variant_level=1
                    )

                self.EmitSignalDetailed(
                    "",
                    "Activated",
                    "oa{sv}",
                    [dbus.ObjectPath(session_handle), opts],
                    details={"destination": sender},
                )

            GLib.timeout_add(self.activated_after, send_activated)

            if self.deactivated_after > 0:

                def send_deactivated():
                    opts = {}
                    if current_eis_serial:
                        opts["activation_id"] = dbus.UInt32(
                            current_eis_serial, variant_level=1
                        )

                    if self.deactivated_position is not None:
                        opts["cursor_position"] = dbus.Struct(
                            self.deactivated_position, signature="dd", variant_level=1
                        )

                    self.EmitSignalDetailed(
                        "",
                        "Deactivated",
                        "oa{sv}",
                        [dbus.ObjectPath(session_handle), opts],
                        details={"destination": sender},
                    )

                GLib.timeout_add(
                    self.activated_after + self.deactivated_after, send_deactivated
                )

        if self.disabled_after > 0:

            def send_disabled():
                self.EmitSignalDetailed(
                    "",
                    "Disabled",
                    "oa{sv}",
                    [dbus.ObjectPath(session_handle), {}],
                    details={"destination": sender},
                )

            GLib.timeout_add(self.disabled_after, send_disabled)

        if self.close_after_enable > 0:
            session = self.active_sessions[session_handle]
            session.close({}, self.close_after_enable)

    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}",
    out_signature="",
)
def Disable(self, session_handle, options, sender):
    try:
        logger.debug(f"Disable with options {options}")
        allowed_options = []

        if not all([k in allowed_options for k in options]):
            logger.error("Disable does not support options")
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}",
    out_signature="",
)
def Release(self, session_handle, options, sender):
    try:
        logger.debug(f"Release with options {options}")
        allowed_options = ["cursor_position"]

        if not all([k in allowed_options for k in options]):
            logger.error("Invalid options for Release")
    except Exception as e:
        logger.critical(e)



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from . import PortalTest

import dbus
import gi
import logging
import os

from typing import NamedTuple, TextIO

gi.require_version("Xdp", "1.0")
from gi.repository import GLib, Xdp

logger = logging.getLogger(__name__)


class SessionCreationFailed(Exception):
    def __init__(self, glib_error):
        self.glib_error = glib_error

    def __str__(self):
        return f"{type(self).__name__}: {self.glib_error}"


class SessionSetup(NamedTuple):
    session: Xdp.Session
    session_handle_token: str
    pw_fd: TextIO


class TestScreenCast(PortalTest):
    def test_version(self):
        self.assert_version_eq(4)

    def create_session(
        self,
        params=None,
        parent=None,
        cancellable=None,
        outputs=Xdp.OutputType.MONITOR,
        flags=Xdp.ScreencastFlags.NONE,
        cursor_mode=Xdp.CursorMode.HIDDEN,
        persist_mode=Xdp.PersistMode.NONE,
        restore_token=None,
    ) -> SessionSetup:
        extra_templates = [("RemoteDesktop", {})]
        self.setup_daemon(params=params or {}, extra_templates=extra_templates)

        xdp = Xdp.Portal.new()
        assert xdp is not None

        session, session_error = None, None
        create_session_done_invoked = False

        def create_session_done(portal, task, data):
            nonlocal session, session_error
            nonlocal create_session_done_invoked

            create_session_done_invoked = True

            try:
                session = portal.create_screencast_session_finish(task)
            except GLib.GError as e:
                session_error = e
            self.mainloop.quit()

        xdp.create_screencast_session(
            outputs=outputs,
            flags=flags,
            cursor_mode=cursor_mode,
            persist_mode=persist_mode,
            restore_token=restore_token,
            cancellable=cancellable,
            callback=create_session_done,
            data=None,
        )

        self.mainloop.run()
        assert create_session_done_invoked
        if session_error is not None:
            raise SessionCreationFailed(session_error)

        # Extract our expected session id. This isn't available from
        # XdpSession so we need to go around it. We can't easily get the
        # sender id so the full path is hard. Let's just extract the token and
        # pretend that's good enough.
        method_calls = self.mock_interface.GetMethodCalls("CreateSession")
        assert len(method_calls) >= 1
        _, args = method_calls.pop()  # Assume the latest has our session
        (options,) = args
        session_handle = options["session_handle_token"]

        assert session.get_session_type() == Xdp.SessionType.SCREENCAST

        # open the PW remote by default since we need it anyway for Start()
        handle = session.open_pipewire_remote()
        pw_fd = os.fdopen(handle)
        assert pw_fd.read() == "I AM GROO^WPIPEWIRE"

        return SessionSetup(
            session=session,
            session_handle_token=session_handle,
            pw_fd=pw_fd,
        )

    def test_create_session(self):
        """
        Create a session with some "random" values and ensure that they're
        passed through to the portal.
        """
        outputs = Xdp.OutputType.MONITOR | Xdp.OutputType.WINDOW
        cursor_mode = Xdp.CursorMode.METADATA
        persist_mode = Xdp.PersistMode.PERSISTENT
        restore_token = "12345"
        flags = Xdp.ScreencastFlags.MULTIPLE

        self.create_session(
            outputs=outputs,
            flags=flags,
            cursor_mode=cursor_mode,
            restore_token=restore_token,
            persist_mode=persist_mode,
        )

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("CreateSession")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        (options,) = args

        assert list(options.keys()) == [
            "handle_token",
            "session_handle_token",
        ]

        # This method doesn't exist on the interface anyway but since we share
        # code paths with RemoteDesktop let's ensure we didn't event try to
        # call it.
        method_calls = self.mock_interface.GetMethodCalls("SelectDevices")
        assert len(method_calls) == 0

        method_calls = self.mock_interface.GetMethodCalls("SelectSources")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args

        assert list(options.keys()) == [
            "handle_token",
            "types",
            "multiple",
            "cursor_mode",
            "persist_mode",
            "restore_token",
        ]

        assert options["types"] == outputs
        assert options["multiple"] == flags
        assert options["cursor_mode"] == cursor_mode
        assert options["restore_token"] == restore_token
        assert options["persist_mode"] == persist_mode

        method_calls = self.mock_interface.GetMethodCalls("OpenPipeWireRemote")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args
        assert list(options.keys()) == []

    def test_create_session_v3(self):
        """
        persist_mode and restore_token were added in v4 of the interface, must
        not be part of SelectSources
        """
        params = {"version": 3}

        persist_mode = Xdp.PersistMode.PERSISTENT
        restore_token = "12345"
        self.create_session(
            params=params,
            restore_token=restore_token,
            persist_mode=persist_mode,
        )

        method_calls = self.mock_interface.GetMethodCalls("SelectSources")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args

        assert "persist_mode" not in options.keys()
        assert "restore_token" not in options.keys()

    def test_session_restore_token_null_if_not_started(self):
        """
        Passed in a restore token but unless the session is started it
        should be NULL.
        """
        restore_token = "12345"
        persist_mode = Xdp.PersistMode.TRANSIENT
        params = {
            "persist-mode": persist_mode,
        }
        setup = self.create_session(
            params=params, restore_token=restore_token, persist_mode=persist_mode
        )

        session = setup.session
        assert session.get_restore_token() is None

    def test_session_start(self):
        """
        Create and start a session
        """

        s = [
            {
                "position": (1, 2),
                "size": (32, 43),
                "source_type": dbus.UInt32(Xdp.OutputType.WINDOW),
            },
            {
                "position": (3, 4),
                "size": (54, 65),
                "source_type": dbus.UInt32(Xdp.OutputType.MONITOR),
            },
        ]

        streams = dbus.Array(
            [(dbus.UInt32(idx), stream) for idx, stream in enumerate(s)],
            signature="(ua{sv})",
            variant_level=1,
        )
        params = {
            "streams": streams,
            "persist-mode": Xdp.PersistMode.TRANSIENT,
        }

        restore_token = "foo"  # ignored since we're not continuing a session
        persist_mode = Xdp.PersistMode.PERSISTENT
        setup = self.create_session(
            params=params, restore_token=restore_token, persist_mode=persist_mode
        )

        session = setup.session
        start_done_invoked = False
        start_result, start_error = None, None

        def start_done(portal, task, data):
            nonlocal start_done_invoked
            nonlocal start_result, start_error

            start_done_invoked = True
            try:
                start_result = session.start_finish(task)
            except GLib.Error as e:
                start_error = e
            self.mainloop.quit()

        session.start(parent=None, cancellable=None, callback=start_done, data=None)

        self.mainloop.run()
        assert start_done_invoked
        assert start_error is None
        assert start_result is True

        # restore_tokenN is what the template uses
        assert session.get_restore_token() == "restore_token0"

        # FIXME: get_streams() returns a GVariant but how do you dereference that?
        session_streams = session.get_streams()
        for s_in, s_out in zip(session_streams, streams):
            assert s_in == s_out

        # We wanted persistent but we'll get transient from the template
        assert session.get_persist_mode() == Xdp.PersistMode.TRANSIENT

        method_calls = self.mock_interface.GetMethodCalls("Start")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, parent_window, options = args
        assert parent_window == ""
        assert list(options.keys()) == ["handle_token"]

    def test_close_session(self):
        """
        Ensure that closing our session explicitly closes the session on DBus
        """
        setup = self.create_session()
        session = setup.session

        was_closed = False

        def method_called(method_name, method_args, path):
            nonlocal was_closed

            if method_name == "Close" and path.endswith(setup.session_handle_token):
                was_closed = True
                self.mainloop.quit()

        bus = self.get_dbus()
        bus.add_signal_receiver(
            handler_function=method_called,
            signal_name="MethodCalled",
            dbus_interface="org.freedesktop.DBus.Mock",
            path_keyword="path",
        )

        session.close()
        self.mainloop.run()
        assert was_closed is True



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from . import PortalTest

import gi
import logging

gi.require_version("Xdp", "1.0")
from gi.repository import GLib, Gio, GioUnix, Xdp

logger = logging.getLogger(__name__)

SVG_IMAGE_DATA = (
    '<?xml version="1.0" encoding="UTF-8"?>'
    '<svg xmlns="http://www.w3.org/2000/svg" height="16px" width="16px"/>'
)


class TestNotification(PortalTest):
    def test_version(self):
        self.assert_version_eq(2)

    def add_notification(
        self, version: int, id_to_set: str, notification_to_set: GLib.Variant
    ):
        params = {"version": version}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        assert xdp is not None

        notification_added = False

        def add_notification_done(portal, task, data):
            nonlocal notification_added
            notification_added = portal.add_notification_finish(task)
            self.mainloop.quit()

        xdp.add_notification(
            id=id_to_set,
            notification=notification_to_set,
            flags=Xdp.NotificationFlags.NONE,
            cancellable=None,
            callback=add_notification_done,
            data=None,
        )

        self.mainloop.run()

        method_calls = self.mock_interface.GetMethodCalls("AddNotification")
        assert len(method_calls) == 1
        timestamp, args = method_calls.pop(0)
        id, notification = args
        assert id == id_to_set
        assert notification_added

        return notification

    def test_add_notification(self):
        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(2, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        notification_to_set_dict.pop("invalid-property", None)
        assert dict(notification) == notification_to_set_dict

    def test_add_notification_bytes_icon(self):
        bytes = GLib.Bytes.new(SVG_IMAGE_DATA.encode("utf-8"))
        icon = Gio.BytesIcon.new(bytes)

        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "icon": icon.serialize(),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(2, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        icon_to_set = notification_to_set_dict.pop("icon", None)
        icon = notification.pop("icon", None)
        assert dict(notification) == notification_to_set_dict
        (icon_type, handle) = icon
        assert icon_type == "file-descriptor"
        input_stream = GioUnix.InputStream.new(handle.take(), True)

        res_bytes = input_stream.read_bytes(bytes.get_size(), None)
        assert res_bytes.equal(bytes)

    def test_add_notification_file_icon(self):
        bytes = GLib.Bytes.new(SVG_IMAGE_DATA.encode("utf-8"))

        (file, iostream) = Gio.File.new_tmp("iconXXXXXX")
        stream = iostream.get_output_stream()
        stream.write_bytes(bytes)
        stream.close()
        icon = Gio.FileIcon.new(file)

        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "icon": icon.serialize(),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(2, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        icon_to_set = notification_to_set_dict.pop("icon", None)
        icon = notification.pop("icon", None)
        assert dict(notification) == notification_to_set_dict
        (icon_type, handle) = icon
        assert icon_type == "file-descriptor"
        input_stream = GioUnix.InputStream.new(handle.take(), True)

        res_bytes = input_stream.read_bytes(bytes.get_size(), None)
        assert res_bytes.equal(bytes)

        file.delete()

    def test_add_notification_themed_icon(self):
        icon = Gio.ThemedIcon.new("someicon")

        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "icon": icon.serialize(),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(2, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        icon_to_set = notification_to_set_dict.pop("icon", None)
        icon_out = notification.pop("icon", None)
        assert dict(notification) == notification_to_set_dict
        (icon_type, names) = icon_out
        assert icon_type == "themed"

        assert names == icon.get_names()

    def test_add_notification_bytes_sound(self):
        bytes = GLib.Bytes.new(SVG_IMAGE_DATA.encode("utf-8"))
        icon = Gio.BytesIcon.new(bytes)

        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "sound": GLib.Variant(
                    "(sv)", ("bytes", GLib.Variant("ay", bytes.get_data()))
                ),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(2, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        sound_to_set = notification_to_set_dict.pop("sound", None)
        sound = notification.pop("sound", None)
        assert dict(notification) == notification_to_set_dict
        (sound_type, handle) = sound
        assert sound_type == "file-descriptor"
        input_stream = GioUnix.InputStream.new(handle.take(), True)

        res_bytes = input_stream.read_bytes(bytes.get_size(), None)
        assert res_bytes.equal(bytes)

    def test_add_notification_file_sound(self):
        bytes = GLib.Bytes.new(SVG_IMAGE_DATA.encode("utf-8"))

        (file, iostream) = Gio.File.new_tmp("soundXXXXXX")
        stream = iostream.get_output_stream()
        stream.write_bytes(bytes)
        stream.close()

        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "sound": GLib.Variant(
                    "(sv)", ["file", GLib.Variant("s", file.get_uri())]
                ),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(2, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        sound_to_set = notification_to_set_dict.pop("sound", None)
        sound_out = notification.pop("sound", None)
        assert dict(notification) == notification_to_set_dict
        (sound_type, handle) = sound_out
        assert sound_type == "file-descriptor"
        input_stream = GioUnix.InputStream.new(handle.take(), True)

        res_bytes = input_stream.read_bytes(bytes.get_size(), None)
        assert res_bytes.equal(bytes)

        file.delete()

    def test_add_notification_v1(self):
        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "invalid-property": GLib.Variant("s", "invalid"),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )

        notification = self.add_notification(1, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        notification_to_set_dict.pop("invalid-property", None)
        assert dict(notification) == notification_to_set_dict

    def test_add_notification_bytes_icon_v1(self):
        bytes = GLib.Bytes.new(SVG_IMAGE_DATA.encode("utf-8"))
        icon = Gio.BytesIcon.new(bytes)

        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "icon": icon.serialize(),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(1, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        icon_to_set = notification_to_set_dict.pop("icon", None)
        icon = notification.pop("icon", None)
        assert dict(notification) == notification_to_set_dict
        (icon_type, res_bytes) = icon
        assert icon_type == "bytes"
        assert GLib.Bytes(res_bytes).equal(bytes)

    def test_add_notification_file_icon_v1(self):
        bytes = GLib.Bytes.new(SVG_IMAGE_DATA.encode("utf-8"))

        (file, iostream) = Gio.File.new_tmp("iconXXXXXX")
        stream = iostream.get_output_stream()
        stream.write_bytes(bytes)
        stream.close()
        icon = Gio.FileIcon.new(file)

        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "icon": icon.serialize(),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(1, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        icon_to_set = notification_to_set_dict.pop("icon", None)
        icon = notification.pop("icon", None)
        assert dict(notification) == notification_to_set_dict
        (icon_type, res_bytes) = icon
        assert icon_type == "bytes"
        assert GLib.Bytes(res_bytes).equal(bytes)

        file.delete()

    def test_add_notification_themed_icon_v1(self):
        icon = Gio.ThemedIcon.new("someicon")

        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "icon": icon.serialize(),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(2, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        icon_to_set = notification_to_set_dict.pop("icon", None)
        icon_out = notification.pop("icon", None)
        assert dict(notification) == notification_to_set_dict
        (icon_type, names) = icon_out
        assert icon_type == "themed"

        assert names == icon.get_names()

    def test_add_notification_buttons(self):
        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "buttons": GLib.Variant(
                    "aa{sv}",
                    [
                        {
                            "label": GLib.Variant("s", "button1"),
                            "action": GLib.Variant("s", "action1"),
                            "purpose": GLib.Variant("s", "some-purpose"),
                        },
                        {
                            "label": GLib.Variant("s", "button2"),
                            "action": GLib.Variant("s", "action2"),
                            "purpose": GLib.Variant("s", "some-purpose"),
                        },
                    ],
                ),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )
        notification = self.add_notification(2, "someid", notification_to_set)

        assert dict(notification) == dict(notification_to_set)

    def test_add_notification_buttons_v1(self):
        notification_to_set = GLib.Variant(
            "a{sv}",
            {
                "title": GLib.Variant("s", "title"),
                "body": GLib.Variant("s", "test notification body"),
                "priority": GLib.Variant("s", "normal"),
                "buttons": GLib.Variant(
                    "aa{sv}",
                    [
                        {
                            "label": GLib.Variant("s", "button1"),
                            "action": GLib.Variant("s", "action1"),
                            "purpose": GLib.Variant("s", "some-purpose"),
                        },
                        {
                            "label": GLib.Variant("s", "button2"),
                            "action": GLib.Variant("s", "action2"),
                            "purpose": GLib.Variant("s", "some-purpose"),
                        },
                    ],
                ),
                "default-action": GLib.Variant("s", "test-action"),
            },
        )

        notification = self.add_notification(1, "someid", notification_to_set)

        notification_to_set_dict = dict(notification_to_set)
        buttons = notification_to_set_dict.pop("buttons", None)

        notification_to_set_dict["buttons"] = []
        for button in buttons:
            button.pop("purpose", None)
            notification_to_set_dict["buttons"].append(button)

        assert dict(notification) == notification_to_set_dict



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from . import PortalTest

import gc
import gi
import logging
import os

from typing import NamedTuple, TextIO

gi.require_version("Xdp", "1.0")
from gi.repository import Gio, GLib, Xdp

logger = logging.getLogger(__name__)


class SessionCreationFailed(Exception):
    def __init__(self, glib_error):
        self.glib_error = glib_error

    def __str__(self):
        return f"{type(self).__name__}: {self.glib_error}"


class SessionSetup(NamedTuple):
    session: Xdp.Session
    session_handle_token: str
    pw_fd: TextIO


class TestRemoteDesktop(PortalTest):
    def test_version(self):
        self.assert_version_eq(2)

    def short_mainloop(self):
        """
        Run a short iteration of our mainloop. Useful to for pure method calls
        that still require dbusmock etc. to update after the invocation.
        """
        GLib.timeout_add(200, self.mainloop.quit)
        self.mainloop.run()

    def create_session(
        self,
        params=None,
        parent=None,
        cancellable=None,
        devices=Xdp.DeviceType.POINTER
        | Xdp.DeviceType.KEYBOARD
        | Xdp.DeviceType.TOUCHSCREEN,
        outputs=Xdp.OutputType.NONE,
        flags=Xdp.RemoteDesktopFlags.NONE,
        cursor_mode=Xdp.CursorMode.HIDDEN,
        start_session=True,
        persist_mode=None,
        restore_token=None,
        cb_closed_signal=None,
    ) -> SessionSetup:
        params = params or {}
        # To make the tests easier, load ScreenCast automatically if we have
        # any outputs specified.
        extra_templates = [("ScreenCast", {})]
        self.setup_daemon(params=params, extra_templates=extra_templates)

        xdp = Xdp.Portal.new()
        assert xdp is not None

        session, session_error = None, None
        create_session_done_invoked = False

        def create_session_done(portal, task, data):
            nonlocal session, session_error
            nonlocal create_session_done_invoked

            create_session_done_invoked = True

            try:
                session = portal.create_remote_desktop_session_finish(task)
            except GLib.GError as e:
                session_error = e
            self.mainloop.quit()

        if restore_token is not None and persist_mode is not None:
            xdp.create_remote_desktop_session_full(
                devices=devices,
                outputs=outputs,
                flags=flags,
                cursor_mode=cursor_mode,
                persist_mode=persist_mode,
                restore_token=restore_token,
                cancellable=cancellable,
                callback=create_session_done,
                data=None,
            )
        else:
            xdp.create_remote_desktop_session(
                devices=devices,
                outputs=outputs,
                flags=flags,
                cursor_mode=cursor_mode,
                cancellable=cancellable,
                callback=create_session_done,
                data=None,
            )
        self.mainloop.run()
        assert create_session_done_invoked
        if session_error is not None:
            raise SessionCreationFailed(session_error)

        # Extract our expected session id. This isn't available from
        # XdpSession so we need to go around it. We can't easily get the
        # sender id so the full path is hard. Let's just extract the token and
        # pretend that's good enough.
        method_calls = self.mock_interface.GetMethodCalls("CreateSession")
        assert len(method_calls) >= 1
        _, args = method_calls.pop()  # Assume the latest has our session
        (options,) = args
        session_handle = options["session_handle_token"]

        assert session.get_session_type() == Xdp.SessionType.REMOTE_DESKTOP

        if cb_closed_signal:
            session.connect("closed", cb_closed_signal)

        if outputs:
            # open the PW remote by default since we need it anyway for Start()
            handle = session.open_pipewire_remote()
            pw_fd = os.fdopen(handle)
            assert pw_fd.read() == "I AM GROO^WPIPEWIRE"
        else:
            pw_fd = None

        if start_session:

            def start_done(portal, task):
                session.start_finish(task)
                self.mainloop.quit()

            session.start(callback=start_done)
            self.mainloop.run()

        return SessionSetup(
            session=session,
            session_handle_token=session_handle,
            pw_fd=pw_fd,
        )

    def test_create_session(self):
        """
        Create a session with some "random" values and ensure that they're
        passed through to the portal.
        """
        devices = Xdp.DeviceType.POINTER | Xdp.DeviceType.KEYBOARD
        outputs = Xdp.OutputType.MONITOR | Xdp.OutputType.WINDOW
        cursor_mode = Xdp.CursorMode.METADATA
        flags = Xdp.RemoteDesktopFlags.MULTIPLE

        self.create_session(
            devices=devices,
            outputs=outputs,
            flags=flags,
            cursor_mode=cursor_mode,
            start_session=False,
        )

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("CreateSession")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        (options,) = args

        assert list(options.keys()) == [
            "handle_token",
            "session_handle_token",
        ]

        method_calls = self.mock_interface.GetMethodCalls("SelectDevices")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args

        assert list(options.keys()) == [
            "handle_token",
            "types",
            "persist_mode",
        ]
        assert options["types"] == devices
        assert options["persist_mode"] == Xdp.PersistMode.NONE

        method_calls = self.mock_interface.GetMethodCalls("SelectSources")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args

        assert list(options.keys()) == [
            "handle_token",
            "types",
            "multiple",
            "cursor_mode",
            "persist_mode",
        ]

        assert options["types"] == outputs
        assert options["multiple"] == flags
        assert options["cursor_mode"] == cursor_mode
        assert options["persist_mode"] == Xdp.PersistMode.NONE

    def test_create_session_restore(self):
        """
        Create a session with some "random" values and ensure that they're
        passed through to the portal.
        """
        devices = Xdp.DeviceType.POINTER | Xdp.DeviceType.KEYBOARD
        outputs = Xdp.OutputType.MONITOR | Xdp.OutputType.WINDOW
        cursor_mode = Xdp.CursorMode.METADATA
        flags = Xdp.RemoteDesktopFlags.MULTIPLE
        persist_mode = Xdp.PersistMode.PERSISTENT
        restore_token = "12345"

        self.create_session(
            devices=devices,
            outputs=outputs,
            flags=flags,
            cursor_mode=cursor_mode,
            persist_mode=persist_mode,
            restore_token=restore_token,
            start_session=False,
        )

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("CreateSession")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        (options,) = args

        assert list(options.keys()) == [
            "handle_token",
            "session_handle_token",
        ]

        method_calls = self.mock_interface.GetMethodCalls("SelectDevices")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args

        assert list(options.keys()) == [
            "handle_token",
            "types",
            "persist_mode",
            "restore_token",
        ]
        assert options["types"] == devices
        assert options["persist_mode"] == persist_mode
        assert options["restore_token"] == restore_token

        method_calls = self.mock_interface.GetMethodCalls("SelectSources")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args

        assert list(options.keys()) == [
            "handle_token",
            "types",
            "multiple",
            "cursor_mode",
            "persist_mode",
            "restore_token",
        ]

        assert options["types"] == outputs
        assert options["multiple"] == flags
        assert options["cursor_mode"] == cursor_mode

    def test_create_session_no_outputs(self):
        """
        Create a session with no output value, ensure it's a "pure"
        remote desktop session.
        """
        devices = Xdp.DeviceType.POINTER | Xdp.DeviceType.KEYBOARD
        outputs = Xdp.OutputType.NONE
        cursor_mode = Xdp.CursorMode.METADATA
        flags = Xdp.RemoteDesktopFlags.MULTIPLE

        self.create_session(
            devices=devices,
            outputs=outputs,
            flags=flags,
            cursor_mode=cursor_mode,
            start_session=False,
        )

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("CreateSession")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        (options,) = args

        assert list(options.keys()) == [
            "handle_token",
            "session_handle_token",
        ]

        method_calls = self.mock_interface.GetMethodCalls("SelectDevices")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args

        assert list(options.keys()) == [
            "handle_token",
            "types",
            "persist_mode",
        ]
        assert options["types"] == devices
        assert options["persist_mode"] == Xdp.PersistMode.NONE

        # No outputs means this should never get called
        method_calls = self.mock_interface.GetMethodCalls("SelectSources")
        assert len(method_calls) == 0

    def test_session_start(self):
        """
        Create and start a session
        """
        devices = Xdp.DeviceType.POINTER | Xdp.DeviceType.KEYBOARD
        outputs = Xdp.OutputType.NONE
        params = {"devices": devices}

        setup = self.create_session(
            params=params,
            devices=devices,
            outputs=outputs,
            start_session=False,
        )

        session = setup.session
        start_done_invoked = False
        start_result, start_error = None, None

        def start_done(portal, task):
            nonlocal start_done_invoked
            nonlocal start_result, start_error

            start_done_invoked = True
            try:
                start_result = session.start_finish(task)
            except GLib.Error as e:
                start_error = e
            self.mainloop.quit()

        session.start(parent=None, cancellable=None, callback=start_done)

        self.mainloop.run()
        assert start_done_invoked
        assert start_error is None
        assert start_result is True

        method_calls = self.mock_interface.GetMethodCalls("Start")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, parent_window, options = args
        assert parent_window == ""
        assert list(options.keys()) == ["handle_token"]

        assert session.get_devices() == devices

        # This is a pure RD session, so all of the SC-specific ones are
        # expected to be NULL-like
        assert session.get_persist_mode() == Xdp.PersistMode.NONE
        assert session.get_streams() is None
        assert session.get_restore_token() is None

    def test_notify_button(self):
        setup = self.create_session()
        session = setup.session

        session.pointer_button(2, Xdp.ButtonState.PRESSED)
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyPointerButton")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, button, state = args
        assert button == 2
        assert state == 1

    def test_notify_key(self):
        setup = self.create_session()
        session = setup.session

        session.keyboard_key(False, 3, Xdp.KeyState.PRESSED)
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyKeyboardKeycode")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, key, state = args
        assert key == 3
        assert state == 1

        session.keyboard_key(True, 4, Xdp.KeyState.PRESSED)
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyKeyboardKeysym")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, key, state = args
        assert key == 4
        assert state == 1

    def test_notify_motion(self):
        setup = self.create_session()
        session = setup.session

        session.pointer_motion(-1.4, -10.1)
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyPointerMotion")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, x, y = args
        assert (x, y) == (-1.4, -10.1)

    def test_notify_motion_absolute(self):
        setup = self.create_session()
        session = setup.session

        session.pointer_position(123, -1.3, -5.2)
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyPointerMotionAbsolute")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, stream, x, y = args
        assert stream == 123
        assert (x, y) == (-1.3, -5.2)

    def test_notify_axis(self):
        setup = self.create_session()
        session = setup.session

        session.pointer_axis(False, -3.3, -8.2)
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyPointerAxis")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, x, y = args
        assert "finish" in options
        assert not options["finish"]
        assert (x, y) == (-3.3, -8.2)

    def test_notify_axis_discrete(self):
        setup = self.create_session()
        session = setup.session

        session.pointer_axis_discrete(Xdp.DiscreteAxis.HORIZONTAL_SCROLL, 4)
        session.pointer_axis_discrete(Xdp.DiscreteAxis.VERTICAL_SCROLL, 6)
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyPointerAxisDiscrete")
        assert len(method_calls) == 2
        _, args = method_calls.pop(0)
        session_handle, options, axis, steps = args
        assert axis == Xdp.DiscreteAxis.HORIZONTAL_SCROLL
        assert steps == 4

        _, args = method_calls.pop(0)
        session_handle, options, axis, steps = args
        assert axis == Xdp.DiscreteAxis.VERTICAL_SCROLL
        assert steps == 6

    def test_notify_touch(self):
        setup = self.create_session()
        session = setup.session

        session.touch_down(2, 3, 4.4, 5.5)
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyTouchDown")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, stream, slot, x, y = args
        assert (stream, slot, x, y) == (2, 3, 4.4, 5.5)

        session.touch_position(6, 7, 8.8, 9.9)  # libportal doesn't check touch state
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyTouchMotion")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, stream, slot, x, y = args
        assert (stream, slot, x, y) == (6, 7, 8.8, 9.9)

        session.touch_up(10)
        self.short_mainloop()

        method_calls = self.mock_interface.GetMethodCalls("NotifyTouchUp")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, slot = args
        assert slot == 10

    def test_connect_to_eis_v1(self):
        params = {"version": 1}
        setup = self.create_session(params=params, start_session=True)
        session = setup.session

        try:
            session.connect_to_eis()
            assert False, "Expected an exception here"
        except GLib.GError as err:
            assert err.matches(Gio.io_error_quark(), Gio.IOErrorEnum.NOT_SUPPORTED)

    def test_connect_to_eis(self):
        setup = self.create_session(start_session=True)
        session = setup.session
        handle = session.connect_to_eis()
        assert handle

        fd = os.fdopen(handle)
        buf = fd.read()
        assert buf == "VANILLA"  # template sends this by default

        method_calls = self.mock_interface.GetMethodCalls("ConnectToEIS")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args
        assert "handle_token" not in options  # this is not a Request
        assert list(options.keys()) == []

    def test_connect_to_eis_fail_reconnect(self):
        setup = self.create_session(start_session=True)
        session = setup.session
        handle = session.connect_to_eis()
        assert handle

        try:
            session.connect_to_eis()
            assert False, "Expected an exception here"
        except GLib.GError as err:
            assert err.matches(Gio.io_error_quark(), Gio.IOErrorEnum.INVALID_ARGUMENT)

    def test_connect_to_eis_fail_connect_before_start(self):
        setup = self.create_session(start_session=False)
        session = setup.session
        try:
            session.connect_to_eis()
            assert False, "Expected an exception here"
        except GLib.GError as err:
            assert err.matches(Gio.io_error_quark(), Gio.IOErrorEnum.INVALID_ARGUMENT)

    def test_close_session(self):
        """
        Ensure that closing our session explicitly closes the session on DBus
        """
        setup = self.create_session()
        session = setup.session

        was_closed = False

        def method_called(method_name, method_args, path):
            nonlocal was_closed

            if method_name == "Close" and path.endswith(setup.session_handle_token):
                was_closed = True
                self.mainloop.quit()

        bus = self.get_dbus()
        bus.add_signal_receiver(
            handler_function=method_called,
            signal_name="MethodCalled",
            dbus_interface="org.freedesktop.DBus.Mock",
            path_keyword="path",
        )

        session.close()
        self.mainloop.run()

        assert was_closed is True

    def test_close_session_signal(self):
        """
        Ensure that we get the GObject signal when our session is closed
        externally.
        """

        session_closed_signal_received = False

        def session_closed(session):
            nonlocal session_closed_signal_received
            session_closed_signal_received = True
            self.mainloop.quit()

        params = {"close-after-start": 500}
        _ = self.create_session(
            params=params, cb_closed_signal=session_closed
        )  # keep SessionResult alive
        self.mainloop.run()

        assert session_closed_signal_received is True

    def test_close_session_signal_after_unref(self):
        """
        Ensure that the portal's internal session reference keeps the session
        alive even after the caller drops their reference.
        """

        session_closed_signal_received = False

        def session_closed(session):
            nonlocal session_closed_signal_received
            session_closed_signal_received = True
            self.mainloop.quit()

        params = {"close-after-start": 500}
        setup = self.create_session(params=params, cb_closed_signal=session_closed)

        # Make GLib critical warnings fatal so that a double-unref
        # (g_object_unref on an already-finalized object) aborts the process
        # rather than being silently ignored.
        old_flags = GLib.log_set_always_fatal(
            GLib.LogLevelFlags.LEVEL_CRITICAL | GLib.LogLevelFlags.LEVEL_WARNING
        )

        try:
            # Drop our session reference
            del setup
            gc.collect()

            self.mainloop.run()
        finally:
            GLib.log_set_always_fatal(old_flags)

        assert session_closed_signal_received is True



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from . import PortalTest
from typing import List, Optional

import gc
import gi
import logging
import pytest
import os

gi.require_version("Xdp", "1.0")
from gi.repository import GLib, Gio, Xdp

logger = logging.getLogger(f"test.{__name__}")
logger.setLevel(logging.DEBUG)


class SessionSetup:
    def __init__(
        self,
        session: Xdp.InputCaptureSession = None,
        zones: Optional[List[Xdp.InputCaptureZone]] = None,
        barriers: Optional[List[Xdp.InputCapturePointerBarrier]] = None,
        failed_barriers: Optional[List[Xdp.InputCapturePointerBarrier]] = None,
        session_handle_token: Optional[str] = None,
    ):
        self.session = session
        self.zones = zones or []
        self.barriers = barriers or []
        self.failed_barriers = failed_barriers or []
        self.session_handle_token = session_handle_token


class SessionCreationFailed(Exception):
    def __init__(self, glib_error):
        self.glib_error = glib_error

    def __str__(self):
        return f"SessionCreationFailed: {self.glib_error}"


class TestInputCapture(PortalTest):
    def create_session_with_barriers(
        self,
        params=None,
        parent=None,
        capabilities=Xdp.InputCapability.POINTER,
        barriers=None,
        allow_failed_barriers=False,
        cancellable=None,
        legacy_session=True,
    ) -> SessionSetup:
        """
        Session creation helper. This function creates a session and sets up
        pointer barriers, with defaults for everything.

        If legacy_session is True (the default), uses the legacy
        CreateSession API. If False, uses CreateSession2 + Start.
        """
        params = params or {}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        assert xdp is not None

        session, session_error = None, None
        create_session_done_invoked = False

        if legacy_session:

            def create_session_done(portal, task, data):
                nonlocal session, session_error
                nonlocal create_session_done_invoked

                create_session_done_invoked = True
                try:
                    session = portal.create_input_capture_session_finish(task)
                    if session is None:
                        session_error = Exception("XdpSession is NULL")
                except GLib.GError as e:
                    session_error = e
                self.mainloop.quit()

            xdp.create_input_capture_session(
                parent=parent,
                capabilities=capabilities,
                cancellable=cancellable,
                callback=create_session_done,
                data=None,
            )

        else:

            def create_session2_done(portal, task, data):
                nonlocal session, session_error
                nonlocal create_session_done_invoked

                create_session_done_invoked = True
                try:
                    session = portal.create_input_capture_session2_finish(task)
                    if session is None:
                        session_error = Exception("XdpInputCaptureSession is NULL")
                except GLib.GError as e:
                    session_error = e
                self.mainloop.quit()

            xdp.create_input_capture_session2(
                cancellable=cancellable,
                callback=create_session2_done,
                data=None,
            )

        self.mainloop.run()
        assert create_session_done_invoked
        if session_error is not None:
            raise SessionCreationFailed(session_error)
        assert session is not None
        assert session.get_session().get_session_type() == Xdp.SessionType.INPUT_CAPTURE

        def extract_session_handle(legacy_session: bool):
            # Extract our expected session id. This isn't available from
            # XdpSession so we need to go around it. We can't easily get the
            # sender id so the full path is hard. Let's just extract the token and
            # pretend that's good enough.
            method_calls = self.mock_interface.GetMethodCalls(
                "CreateSession" if legacy_session else "CreateSession2"
            )
            assert len(method_calls) >= 1
            _, args = method_calls.pop()  # Assume the latest has our session
            if legacy_session:
                (_, options) = args
            else:
                (options,) = args
            return options["session_handle_token"]

        session_handle = extract_session_handle(legacy_session)
        if not legacy_session:
            # Start the session (this also triggers GetZones internally)
            self.start_session(session, parent=parent, capabilities=capabilities)

        zones = session.get_zones()

        if barriers is None:
            barriers = [Xdp.InputCapturePointerBarrier(id=1, x1=0, x2=1920, y1=0, y2=0)]

        # Check that we get the notify:is-active for each barrier
        active_barriers = []
        inactive_barriers = []

        def notify_active_cb(barrier, pspec):
            nonlocal active_barriers, inactive_barriers

            if barrier.props.is_active:
                active_barriers.append(barrier)
            else:
                inactive_barriers.append(barrier)

        for b in barriers:
            b.connect("notify::is-active", notify_active_cb)

        failed_barriers = None

        def set_pointer_barriers_done(session, task, data):
            nonlocal session_error, failed_barriers
            nonlocal set_pointer_barriers_done_invoked

            set_pointer_barriers_done_invoked = True
            try:
                failed_barriers = session.set_pointer_barriers_finish(task)
            except GLib.GError as e:
                session_error = e
            self.mainloop.quit()

        set_pointer_barriers_done_invoked = False
        session.set_pointer_barriers(
            barriers=barriers,
            cancellable=None,
            callback=set_pointer_barriers_done,
            data=None,
        )
        self.mainloop.run()

        if session_error is not None:
            raise SessionCreationFailed(session_error)

        assert set_pointer_barriers_done_invoked
        assert sorted(active_barriers + inactive_barriers) == sorted(barriers)

        if not allow_failed_barriers:
            assert failed_barriers == [], (
                "Barriers failed but allow_failed_barriers was not set"
            )

        return SessionSetup(
            session=session,
            zones=zones,
            barriers=active_barriers,
            failed_barriers=failed_barriers,
            session_handle_token=session_handle,
        )

    def start_session(
        self, session, parent=None, capabilities=Xdp.InputCapability.POINTER
    ):
        """
        Helper to start a session created via CreateSession2. Returns True
        on success.
        """
        start_done_invoked = False
        start_result, start_error = None, None

        def start_done(sess, task, data):
            nonlocal start_done_invoked
            nonlocal start_result, start_error

            start_done_invoked = True
            try:
                start_result = sess.start_finish(task)
            except GLib.GError as e:
                start_error = e
            self.mainloop.quit()

        session.start(
            parent=parent,
            capabilities=capabilities,
            cancellable=None,
            callback=start_done,
            data=None,
        )

        self.mainloop.run()
        assert start_done_invoked
        if start_error is not None:
            raise SessionCreationFailed(start_error)
        return start_result

    def test_version(self):
        """This tests the test suite setup rather than libportal"""
        params = {}
        self.setup_daemon(params)
        assert self.properties_interface.Get(self.INTERFACE_NAME, "version") == 1

    def test_session_create(self):
        """
        The basic test of successful create and zone check
        """
        params = {
            "zones": [(1920, 1080, 0, 0), (1080, 1920, 1920, 1080)],
            "zone-set": 1234,
        }

        capabilities = Xdp.InputCapability.POINTER | Xdp.InputCapability.KEYBOARD

        setup = self.create_session_with_barriers(params, capabilities=capabilities)
        assert setup.session is not None
        zones = setup.zones
        assert len(zones) == 2
        z1 = zones[0]
        assert z1.props.width == 1920
        assert z1.props.height == 1080
        assert z1.props.x == 0
        assert z1.props.y == 0
        assert z1.props.zone_set == 1234

        z2 = zones[1]
        assert z2.props.width == 1080
        assert z2.props.height == 1920
        assert z2.props.x == 1920
        assert z2.props.y == 1080
        assert z2.props.zone_set == 1234

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("CreateSession")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        parent, options = args
        assert list(options.keys()) == [
            "handle_token",
            "session_handle_token",
            "capabilities",
        ]
        assert options["capabilities"] == capabilities

        method_calls = self.mock_interface.GetMethodCalls("GetZones")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args
        assert list(options.keys()) == ["handle_token"]

    def test_session_create_cancel_during_create(self):
        """
        Create a session but cancel while waiting for the CreateSession request
        """
        params = {"delay": 1000}
        cancellable = Gio.Cancellable()
        GLib.timeout_add(300, cancellable.cancel)

        with pytest.raises(SessionCreationFailed) as e:
            self.create_session_with_barriers(params=params, cancellable=cancellable)
            assert "Operation was cancelled" in e.glib_error.message

    def test_session_create_cancel_during_getzones(self):
        """
        Create a session but cancel while waiting for the GetZones request
        """
        # libportal issues two requests: CreateSession and GetZones,
        # param is set for each to delay 500 ms so if we cancel after 700, the
        # one that is cancelled should be the GetZones one.
        # Can't guarantee it but this is the best we can do
        params = {"delay": 500}
        cancellable = Gio.Cancellable()
        GLib.timeout_add(700, cancellable.cancel)

        with pytest.raises(SessionCreationFailed) as e:
            self.create_session_with_barriers(params=params, cancellable=cancellable)
            assert "Operation was cancelled" in e.glib_error.message

    def test_session_create_no_serial_on_getzones(self):
        """
        Test buggy portal implementation not replying with a zone_set in
        GetZones
        """
        params = {
            "zone-set": -1,
        }

        with pytest.raises(SessionCreationFailed):
            self.create_session_with_barriers(params)

    def test_session_create_no_zones_on_getzones(self):
        """
        Test buggy portal implementation not replying with a zone
        GetZones
        """
        params = {
            "zones": [(0, 0, 0, 0)],
        }

        with pytest.raises(SessionCreationFailed):
            self.create_session_with_barriers(params)

    def _test_session_create_without_subref(self):
        """
        Create a new InputCapture session but never access the actual
        input capture session.
        """
        self.setup_daemon({})

        xdp = Xdp.Portal.new()
        assert xdp is not None

        parent_session, session_error = None, None
        create_session_done_invoked = False

        def create_session_done(portal, task, data):
            nonlocal parent_session, session_error
            nonlocal create_session_done_invoked

            create_session_done_invoked = True
            try:
                parent_session = portal.create_input_capture_session_finish(task)
                if parent_session is None:
                    session_error = Exception("XdpSession is NULL")
            except GLib.GError as e:
                session_error = e
            self.mainloop.quit()

        capabilities = Xdp.InputCapability.POINTER | Xdp.InputCapability.KEYBOARD
        xdp.create_input_capture_session(
            parent=None,
            capabilities=capabilities,
            cancellable=None,
            callback=create_session_done,
            data=None,
        )

        self.mainloop.run()
        assert create_session_done_invoked

        # Explicitly don't call parent_session.get_input_capture_session()
        # since that would cause python to g_object_ref the IC session.
        # By not doing so we never ref that object and can test for the correct
        # cleanup

    def _test_connect_to_eis(self, legacy_session):
        params = {}
        setup = self.create_session_with_barriers(params, legacy_session=legacy_session)
        assert setup.session is not None

        handle = setup.session.connect_to_eis()
        assert handle >= 0

        fd = os.fdopen(handle)
        buf = fd.read()
        assert buf == "VANILLA"  # template sends this by default

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("ConnectToEIS")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        parent, options = args
        assert "handle_token" not in options  # This is not a Request
        assert list(options.keys()) == []

    def test_connect_to_eis(self):
        """The basic test of retrieving the EIS handle"""
        self._test_connect_to_eis(legacy_session=True)

    def test_connect_to_eis_session2(self):
        """The basic test of retrieving the EIS handle (session2)"""
        self._test_connect_to_eis(legacy_session=False)

    def _test_pointer_barriers_success(self, legacy_session):
        b1 = Xdp.InputCapturePointerBarrier(id=1, x1=0, x2=1920, y1=0, y2=0)
        b2 = Xdp.InputCapturePointerBarrier(id=2, x1=1920, x2=1920, y1=0, y2=1080)

        params = {}
        setup = self.create_session_with_barriers(
            params, barriers=[b1, b2], legacy_session=legacy_session
        )
        assert setup.barriers == [b1, b2]

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("SetPointerBarriers")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, barriers, zone_set = args
        assert list(options.keys()) == ["handle_token"]
        for b in barriers:
            assert "barrier_id" in b
            assert "position" in b
            assert b["barrier_id"] in [1, 2]
            x1, y1, x2, y2 = [int(x) for x in b["position"]]
            if b["barrier_id"] == 1:
                assert (x1, y1, x2, y2) == (0, 0, 1920, 0)
            if b["barrier_id"] == 2:
                assert (x1, y1, x2, y2) == (1920, 0, 1920, 1080)

    def test_pointer_barriers_success(self):
        """Some successful pointer barriers"""
        self._test_pointer_barriers_success(legacy_session=True)

    def test_pointer_barriers_success_session2(self):
        """Some successful pointer barriers (session2)"""
        self._test_pointer_barriers_success(legacy_session=False)

    def _test_pointer_barriers_failures(self, legacy_session):
        b1 = Xdp.InputCapturePointerBarrier(id=1, x1=0, x2=1920, y1=0, y2=0)
        b2 = Xdp.InputCapturePointerBarrier(id=2, x1=1, x2=2, y1=3, y2=4)
        b3 = Xdp.InputCapturePointerBarrier(id=3, x1=1, x2=2, y1=3, y2=4)
        b4 = Xdp.InputCapturePointerBarrier(id=4, x1=1920, x2=1920, y1=0, y2=1080)

        params = {"failed-barriers": [2, 3]}
        setup = self.create_session_with_barriers(
            params,
            barriers=[b1, b2, b3, b4],
            allow_failed_barriers=True,
            legacy_session=legacy_session,
        )
        assert setup.barriers == [b1, b4]
        assert setup.failed_barriers == [b2, b3]

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("SetPointerBarriers")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options, barriers, zone_set = args
        assert list(options.keys()) == ["handle_token"]
        for b in barriers:
            assert "barrier_id" in b
            assert "position" in b
            assert b["barrier_id"] in [1, 2, 3, 4]
            x1, y1, x2, y2 = [int(x) for x in b["position"]]
            if b["barrier_id"] == 1:
                assert (x1, y1, x2, y2) == (0, 0, 1920, 0)
            if b["barrier_id"] in [2, 3]:
                assert (x1, y1, x2, y2) == (1, 3, 2, 4)
            if b["barrier_id"] == 4:
                assert (x1, y1, x2, y2) == (1920, 0, 1920, 1080)

    def test_pointer_barriers_failures(self):
        """Test with some barriers failing"""
        self._test_pointer_barriers_failures(legacy_session=True)

    def test_pointer_barriers_failures_session2(self):
        """Test with some barriers failing (session2)"""
        self._test_pointer_barriers_failures(legacy_session=False)

    def _test_enable_disable_release(self, legacy_session):
        params = {}

        setup = self.create_session_with_barriers(params, legacy_session=legacy_session)
        session = setup.session

        session.enable()
        session.disable()
        session.release(activation_id=456)  # fake id, doesn't matter here

        self.mainloop.run()

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("Enable")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args
        assert list(options.keys()) == []

        method_calls = self.mock_interface.GetMethodCalls("Disable")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args
        assert list(options.keys()) == []

        method_calls = self.mock_interface.GetMethodCalls("Release")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args
        assert list(options.keys()) == ["activation_id"]

    def test_enable_disable_release(self):
        """Test enable/disable calls"""
        self._test_enable_disable_release(legacy_session=True)

    def test_enable_disable_release_session2(self):
        """Test enable/disable calls (session2)"""
        self._test_enable_disable_release(legacy_session=False)

    def _test_release_at(self, legacy_session):
        params = {}

        setup = self.create_session_with_barriers(params, legacy_session=legacy_session)
        session = setup.session

        # libportal allows us to call Release without Enable first,
        # we just fake an activation_id
        session.release_at(
            activation_id=456, cursor_x_position=10, cursor_y_position=10
        )
        self.mainloop.run()

        # Now verify our DBus calls were correct
        method_calls = self.mock_interface.GetMethodCalls("Release")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, options = args
        assert list(options.keys()) == ["activation_id", "cursor_position"]
        cursor_position = options["cursor_position"]
        assert cursor_position == (10.0, 10.0)

    def test_release_at(self):
        """Test the release_at call with a cursor position"""
        self._test_release_at(legacy_session=True)

    def test_release_at_session2(self):
        """Test the release_at call with a cursor position (session2)"""
        self._test_release_at(legacy_session=False)

    def _test_activated(self, legacy_session):
        params = {
            "eis-serial": 123,
            "activated-after": 20,
            "activated-barrier": 1,
            "activated-position": (10.0, 20.0),
            "deactivated-after": 20,
            "deactivated-position": (20.0, 30.0),
        }

        setup = self.create_session_with_barriers(params, legacy_session=legacy_session)
        session = setup.session

        session_activated_signal_received = False
        session_deactivated_signal_received = False
        signal_activated_options = None
        signal_deactivated_options = None
        signal_activation_id = None
        signal_deactivation_id = None

        def session_activated(session, activation_id, opts):
            nonlocal session_activated_signal_received
            nonlocal signal_activation_id, signal_activated_options
            session_activated_signal_received = True
            signal_activated_options = opts
            signal_activation_id = activation_id

        def session_deactivated(session, activation_id, opts):
            nonlocal session_deactivated_signal_received
            nonlocal signal_deactivation_id, signal_deactivated_options
            session_deactivated_signal_received = True
            signal_deactivated_options = opts
            signal_deactivation_id = activation_id
            self.mainloop.quit()

        session.connect("activated", session_activated)
        session.connect("deactivated", session_deactivated)
        session.enable()

        self.mainloop.run()

        assert session_activated_signal_received
        assert signal_activated_options is not None
        assert signal_activation_id == 123
        assert list(signal_activated_options.keys()) == [
            "activation_id",
            "cursor_position",
            "barrier_id",
        ]
        assert signal_activated_options["barrier_id"] == 1
        assert signal_activated_options["cursor_position"] == (10.0, 20.0)
        assert signal_activated_options["activation_id"] == 123

        assert session_deactivated_signal_received
        assert signal_deactivated_options is not None
        assert signal_deactivation_id == 123
        assert list(signal_deactivated_options.keys()) == [
            "activation_id",
            "cursor_position",
        ]
        assert signal_deactivated_options["cursor_position"] == (20.0, 30.0)
        assert signal_deactivated_options["activation_id"] == 123

    def test_activated(self):
        """Test the Activated signal"""
        self._test_activated(legacy_session=True)

    def test_activated_session2(self):
        """Test the Activated signal (session2)"""
        self._test_activated(legacy_session=False)

    def _test_zones_changed(self, legacy_session):
        params = {
            "zones": [(1920, 1080, 0, 0), (1080, 1920, 1920, 1080)],
            "changed-zones": [(1024, 768, 0, 0)],
            "change-zones-after": 200,
            "zone-set": 567,
        }

        setup = self.create_session_with_barriers(params, legacy_session=legacy_session)
        session = setup.session

        signal_received = False
        signal_options = None
        zone_props = {z: None for z in setup.zones}

        def zones_changed(session, opts):
            nonlocal signal_received, signal_options, zone_props
            signal_received = True
            signal_options = opts
            if signal_received and all([v == False for v in zone_props.values()]):
                self.mainloop.quit()

        session.connect("zones-changed", zones_changed)

        def zones_is_valid_changed(zone, pspec):
            nonlocal zone_props, signal_received
            zone_props[zone] = zone.props.is_valid
            if signal_received and all([v == False for v in zone_props.values()]):
                self.mainloop.quit()

        for z in setup.zones:
            z.connect("notify::is-valid", zones_is_valid_changed)

        self.mainloop.run()

        assert signal_received
        assert signal_options is not None
        assert list(signal_options.keys()) == ["zone_set"]
        assert signal_options["zone_set"] == 567

        assert all([z.props.zone_set == 568 for z in session.get_zones()])
        assert all([v == False for v in zone_props.values()])

    def test_zones_changed(self):
        """Test the ZonesChanged signal"""
        self._test_zones_changed(legacy_session=True)

    def test_zones_changed_session2(self):
        """Test the ZonesChanged signal (session2)"""
        self._test_zones_changed(legacy_session=False)

    def _test_disabled(self, legacy_session):
        params = {
            "disabled-after": 20,
        }

        setup = self.create_session_with_barriers(params, legacy_session=legacy_session)
        session = setup.session

        disabled_signal_received = False

        def session_disabled(session, options):
            nonlocal disabled_signal_received
            disabled_signal_received = True
            self.mainloop.quit()

        session.connect("disabled", session_disabled)

        session.enable()

        self.mainloop.run()

        assert disabled_signal_received

    def test_disabled(self):
        """Test the Disabled signal"""
        self._test_disabled(legacy_session=True)

    def test_disabled_session2(self):
        """Test the Disabled signal (session2)"""
        self._test_disabled(legacy_session=False)

    def _test_close_session(self, legacy_session):
        setup = self.create_session_with_barriers(legacy_session=legacy_session)
        session = setup.session
        xdp_session = setup.session.get_session()

        was_closed = False

        def method_called(method_name, method_args, path):
            nonlocal was_closed

            if method_name == "Close" and path.endswith(setup.session_handle_token):
                was_closed = True
                self.mainloop.quit()

        bus = self.get_dbus()
        bus.add_signal_receiver(
            handler_function=method_called,
            signal_name="MethodCalled",
            dbus_interface="org.freedesktop.DBus.Mock",
            path_keyword="path",
        )

        xdp_session.close()
        self.mainloop.run()

        assert was_closed is True

    def test_close_session(self):
        """Ensure that closing our session explicitly closes the session on DBus."""
        self._test_close_session(legacy_session=True)

    def test_close_session_session2(self):
        """Ensure that closing our session explicitly closes the session on DBus (session2)."""
        self._test_close_session(legacy_session=False)

    def _test_close_session_signal(self, legacy_session):
        params = {"close-after-enable": 500}
        setup = self.create_session_with_barriers(params, legacy_session=legacy_session)
        session = setup.session
        xdp_session = setup.session.get_session()

        session_closed_signal_received = False

        def session_closed(session):
            nonlocal session_closed_signal_received
            session_closed_signal_received = True
            self.mainloop.quit()

        xdp_session.connect("closed", session_closed)

        session.enable()
        self.mainloop.run()

        assert session_closed_signal_received is True

    def test_close_session_signal(self):
        """Ensure that we get the GObject signal when our session is closed externally."""
        self._test_close_session_signal(legacy_session=True)

    def test_close_session_signal_session2(self):
        """Ensure that we get the GObject signal when our session is closed externally (session2)."""
        self._test_close_session_signal(legacy_session=False)

    def _test_close_session_signal_after_unref(self, legacy_session):
        """
        Ensure that the portal's internal session reference keeps the session
        alive even after the caller drops their reference.
        """
        params = {"close-after-enable": 500}
        setup = self.create_session_with_barriers(params, legacy_session=legacy_session)
        session = setup.session
        xdp_session = session.get_session()

        session_closed_signal_received = False

        def session_closed(session):
            nonlocal session_closed_signal_received
            session_closed_signal_received = True
            self.mainloop.quit()

        xdp_session.connect("closed", session_closed)

        session.enable()

        # Drop our session reference
        del setup, session, xdp_session
        gc.collect()

        # Make GLib critical warnings fatal so that a double-unref
        # (g_object_unref on an already-finalized object) aborts the process
        # rather than being silently ignored. This must be set after the del
        # above: the XdpInputCaptureSession uses a weak ref to XdpSession and
        # finalization order may trigger unrelated g_critical messages.
        old_flags = GLib.log_set_always_fatal(
            GLib.LogLevelFlags.LEVEL_CRITICAL | GLib.LogLevelFlags.LEVEL_WARNING
        )

        try:
            self.mainloop.run()
        finally:
            GLib.log_set_always_fatal(old_flags)

        assert session_closed_signal_received is True

    def test_close_session_signal_after_unref(self):
        """Ensure session survives caller unref and receives the D-Bus Closed signal."""
        self._test_close_session_signal_after_unref(legacy_session=True)

    def test_close_session_signal_after_unref_session2(self):
        """Ensure session survives caller unref and receives the D-Bus Closed signal (session2)."""
        self._test_close_session_signal_after_unref(legacy_session=False)

    def test_create_session2(self):
        """
        Test basic CreateSession2 session creation and Start
        """
        params = {}
        setup = self.create_session_with_barriers(params, legacy_session=False)
        assert setup.session is not None

        method_calls = self.mock_interface.GetMethodCalls("CreateSession2")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        (options,) = args
        assert "session_handle_token" in options

        method_calls = self.mock_interface.GetMethodCalls("Start")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, parent_window, options = args
        assert parent_window == ""
        assert "handle_token" in options
        assert "capabilities" in options
        assert options["capabilities"] == Xdp.InputCapability.POINTER

    def test_start_with_persist_mode(self):
        """
        Test that persist_mode is sent in Start options when set
        """
        params = {"start-persist-mode": 2}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        session = xdp.create_input_capture_session2_sync(cancellable=None)

        session.set_session_persistence(Xdp.InputCaptureSessionPersistence.PERSISTENT)

        result = self.start_session(session)
        assert result is True

        method_calls = self.mock_interface.GetMethodCalls("Start")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, parent_window, options = args
        assert "persist_mode" in options
        assert options["persist_mode"] == Xdp.InputCaptureSessionPersistence.PERSISTENT

    def test_start_no_persist_mode_when_none(self):
        """
        Test that persist_mode is NOT sent in Start options when persistence
        is NONE (the default)
        """
        params = {}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        session = xdp.create_input_capture_session2_sync(cancellable=None)

        result = self.start_session(session)
        assert result is True

        method_calls = self.mock_interface.GetMethodCalls("Start")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, parent_window, options = args
        assert "persist_mode" not in options

    def test_start_with_restore_token_v2(self):
        """
        Test that restore_token is sent in Start options when version >= 2
        and a restore token is set
        """
        params = {"version": 2, "start-persist-mode": 2}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        session = xdp.create_input_capture_session2_sync(cancellable=None)

        session.set_session_persistence(Xdp.InputCaptureSessionPersistence.PERSISTENT)
        session.set_restore_token("previous-session-token")

        result = self.start_session(session)
        assert result is True

        method_calls = self.mock_interface.GetMethodCalls("Start")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, parent_window, options = args
        assert "restore_token" in options
        assert options["restore_token"] == "previous-session-token"

    def test_start_no_restore_token_v1(self):
        """
        Test that restore_token is NOT sent in Start options when version == 1,
        even if a restore token is set
        """
        params = {"version": 1}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        session = xdp.create_input_capture_session2_sync(cancellable=None)

        session.set_session_persistence(Xdp.InputCaptureSessionPersistence.PERSISTENT)
        session.set_restore_token("previous-session-token")

        result = self.start_session(session)
        assert result is True

        method_calls = self.mock_interface.GetMethodCalls("Start")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        session_handle, parent_window, options = args
        assert "restore_token" not in options

    def test_start_restore_token_from_response(self):
        """
        Test that the restore token returned by Start is stored on the session
        """
        params = {"version": 2, "start-persist-mode": 2}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        session = xdp.create_input_capture_session2_sync(cancellable=None)

        session.set_session_persistence(Xdp.InputCaptureSessionPersistence.PERSISTENT)

        result = self.start_session(session)
        assert result is True

        # The template returns "restore_tokenN"
        restore_token = session.get_restore_token()
        assert restore_token is not None
        assert restore_token.startswith("restore_token")

    def test_start_no_restore_token_in_response_when_no_persist(self):
        """
        Test that no restore token is returned when persist mode is 0
        """
        params = {"version": 2, "start-persist-mode": 0}
        self.setup_daemon(params)

        xdp = Xdp.Portal.new()
        session = xdp.create_input_capture_session2_sync(cancellable=None)

        result = self.start_session(session)
        assert result is True

        assert session.get_restore_token() is None



# SPDX-License-Identifier: LGPL-3.0-only
#
# This file is formatted with Python Black

from dbusmock import DBusMockObject
from typing import Dict, Any, NamedTuple, Optional
from itertools import count
from gi.repository import GLib

import dbus
import logging


ASVType = Dict[str, Any]

logging.basicConfig(format="%(levelname).1s|%(name)s: %(message)s", level=logging.DEBUG)
logger = logging.getLogger("templates")


class MockParams:
    """
    Helper class for storing template parameters. The Mock object passed into
    ``load()`` is shared between all templates. This makes it easy to have
    per-template parameters by calling:

        >>> params = MockParams.get(mock, MAIN_IFACE)
        >>> params.version = 1

    and later, inside a DBus method:
        >>> params = MockParams.get(self, MAIN_IFACE)
        >>> return params.version
    """

    @classmethod
    def get(cls, mock, interface_name):
        params = getattr(mock, "params", {})
        try:
            return params[interface_name]
        except KeyError:
            c = cls()
            params[interface_name] = c
            mock.params = params
            return c


class Response(NamedTuple):
    response: int
    results: ASVType


class Request:
    _token_counter = count()

    def __init__(
        self, bus_name: dbus.service.BusName, sender: str, options: Optional[ASVType]
    ):
        options = options or {}
        sender_token = sender.removeprefix(":").replace(".", "_")
        handle_token = options.get("handle_token", next(self._token_counter))
        self.sender = sender
        self.handle = (
            f"/org/freedesktop/portal/desktop/request/{sender_token}/{handle_token}"
        )
        self.mock = DBusMockObject(
            bus_name=bus_name,
            path=self.handle,
            interface="org.freedesktop.portal.Request",
            props={},
        )
        self.mock.AddMethod("", "Close", "", "", "self.RemoveObject(self.path)")
        logger.debug(f"Request created at {self.handle}")

    def respond(self, response: Response, delay: int = 0):
        def respond():
            logger.debug(f"Request.Response on {self.handle}: {response}")
            self.mock.EmitSignalDetailed(
                "",
                "Response",
                "ua{sv}",
                [dbus.UInt32(response.response), response.results],
                details={"destination": self.sender},
            )

        if delay > 0:
            GLib.timeout_add(delay, respond)
        else:
            respond()


class Session:
    _token_counter = count()

    def __init__(
        self, bus_name: dbus.service.BusName, sender: str, options: Optional[ASVType]
    ):
        options = options or {}
        sender_token = sender.removeprefix(":").replace(".", "_")
        handle_token = options.get("session_handle_token", next(self._token_counter))
        self.sender = sender
        self.handle = (
            f"/org/freedesktop/portal/desktop/session/{sender_token}/{handle_token}"
        )
        self.mock = DBusMockObject(
            bus_name=bus_name,
            path=self.handle,
            interface="org.freedesktop.portal.Session",
            props={},
        )
        self.mock.AddMethod("", "Close", "", "", "self.RemoveObject(self.path)")
        logger.debug(f"Session created at {self.handle}")

    def close(self, details: ASVType, delay: int = 0):
        def respond():
            logger.debug(f"Session.Closed on {self.handle}: {details}")
            self.mock.EmitSignalDetailed(
                "", "Closed", "a{sv}", [details], details={"destination": self.sender}
            )

        if delay > 0:
            GLib.timeout_add(delay, respond)
        else:
            respond()
